# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
A tool for modifying, comparing and plotting electronic bandstructures.
"""

import importlib.metadata

from . import kpoints
from . import eigenvals
from . import compare
from . import lattice
from . import plot

__version__ = importlib.metadata.version(__name__.replace(".", "-"))

__all__ = ("kpoints", "eigenvals", "compare", "lattice", "plot")



# (c) 2019, Microsoft Research
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines a function to align two bandstructures in such a way
that their difference is minimized.
"""

from collections import namedtuple

import numpy as np
from scipy.optimize import minimize

from fsc.export import export

from .difference import calculate as calculate_diff

AlignResult = namedtuple(
    "AlignResult", ["shift", "eigenvals1_shifted", "eigenvals2_shifted", "difference"]
)


@export
def calculate(eigenvals1, eigenvals2, *, symmetric_shift=False, **kwargs):
    """
    Shift the two sets of eigenvalues such that their difference is minimized.

    :param eigenvals1: The first set of eigenvalues.
    :type eigenvals1: EigenvalsData

    :param eigenvals2: The second set of eigenvalues.
    :type eigenvals2: EigenvalsData

    :param symmetric_shift: If `True`, the two sets of eigenvalues are shifted by
        equal and opposite values. Otherwise, only the second set of eigenvalues
        is shifted.
    :type symmetric_shift: bool

    :param kwargs: Keyword arguments passed on to :func:`.difference.calculate`.
    :type kwargs: dict
    """

    def _do_shift(shift_delta):
        if symmetric_shift:
            return (
                eigenvals1.shift(-shift_delta / 2),
                eigenvals2.shift(shift_delta / 2),
            )
        return eigenvals1, eigenvals2.shift(shift_delta)

    def opt_func(delta):
        (shift_delta,) = delta
        ev1, ev2 = _do_shift(shift_delta=shift_delta)
        return calculate_diff(eigenvals1=ev1, eigenvals2=ev2, **kwargs)

    res = minimize(fun=opt_func, x0=np.array([0]))
    (shift,) = res.x
    ev1, ev2 = _do_shift(shift_delta=shift)
    difference = res.fun
    return AlignResult(
        shift=shift,
        eigenvals1_shifted=ev1,
        eigenvals2_shifted=ev2,
        difference=difference,
    )



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines functions to compare two bandstructures by calculating their difference, with different averaging and weighting methods.
"""

import numpy as np
from fsc.export import export


@export
def calculate(
    eigenvals1,
    eigenvals2,
    *,
    avg_func=np.average,
    weight_eigenval=np.ones_like,
    symmetric_eigenval_weights=True,
    weight_kpoint=lambda kpts: np.ones(np.array(kpts).shape[0])
):
    """
    Calculate the difference between two bandstructures.

    :param eigenvals1: The first set of eigenvalues.
    :type eigenvals1: EigenvalsData

    :param eigenvals2: The second set of eigenvalues.
    :type eigenvals2: EigenvalsData

    :param avg_func: Function which is used to average the difference between the two sets of eigenvalues.

    :param weight_eigenval: A function which takes the eigenvalues as input, and returns the corresponding weights.

    :param symmetric_eigenval_weights: Determines whether both sets of eigenvalues are used to calculate weights, or just the first one.
    :type symmetric_eigenval_weights: bool

    :param weight_kpoint: A function which takes the k-points as input, and returns the corresponding weights.
    """
    kpoints = np.array(eigenvals1.kpoints)
    if not np.allclose(kpoints, np.array(eigenvals2.kpoints)):
        raise ValueError("The k-points of the two sets of eigenvalues do not match.")
    kpoint_weights = weight_kpoint(kpoints)
    if symmetric_eigenval_weights:
        eigenval_weights = np.mean(
            [
                weight_eigenval(eigenvals1.eigenvals),
                weight_eigenval(eigenvals2.eigenvals),
            ],
            axis=0,
        )
    else:
        eigenval_weights = weight_eigenval(eigenvals1.eigenvals)

    weights = (eigenval_weights.T * kpoint_weights).T
    diff = np.abs(eigenvals1.eigenvals - eigenvals2.eigenvals)
    return avg_func(diff, weights=weights)


@export
def energy_window(lower, upper):
    """
    Creates an eigenvalue weighting function that only takes into account eigenvalues in a certain energy window.

    :param lower: Lower bound of the energy window.
    :type lower: float

    :param upper: Upper bound of the energy window.
    :type upper: float
    """
    lower, upper = sorted([lower, upper])

    def weight_eigenval(eigenvals):
        return np.array(np.logical_and(lower < eigenvals, eigenvals < upper), dtype=int)

    return weight_eigenval



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines the data class for a regular k-point mesh.
"""

import types
import itertools

import numpy as np
from fsc.export import export
from fsc.hdf5_io import subscribe_hdf5

from ._base import KpointsBase


@export
@subscribe_hdf5("bands_inspect.kpoints_mesh", extra_tags=("kpoints_mesh",))
class KpointsMesh(KpointsBase, types.SimpleNamespace):
    r"""
    Defines k-points on a regular mesh.

    :param mesh: Defines the grid size (number of different k-point values) for each dimension.
    :type mesh: list

    :param offset: Offset added to the k-point values. If nothing is given, the grid is aligned at the :math:`\Gamma` - point.
    :type offset: list
    """

    def __init__(self, mesh, offset=None):  # pylint: disable=missing-function-docstring
        mesh = tuple(int(m) for m in mesh)
        if offset is None:
            offset = np.zeros_like(mesh)
        else:
            offset = np.array(offset)

        if len(offset) != len(mesh):
            raise ValueError(
                f"Length of 'offset' ({offset}) does not match the length of 'mesh' ({mesh})."
            )
        self.mesh = mesh
        self.offset = offset

    @property
    def kpoints_explicit(self):
        res = np.array(
            list(
                itertools.product(
                    *[np.linspace(0, 1, m, endpoint=False) for m in self.mesh]
                )
            )
        )
        res += self.offset
        res %= 1
        res.flags.writeable = False
        return res

    def to_hdf5(self, hdf5_handle):
        hdf5_handle["mesh"] = self.mesh
        hdf5_handle["offset"] = self.offset

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return cls(mesh=hdf5_handle["mesh"][()], offset=hdf5_handle["offset"][()])



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>

# pylint: disable=undefined-variable
"""
This module contains classes to define sets of k-points, for example for a k-point path, or a regular mesh. All k-points are given in reciprocal lattice coordinates.
"""

from ._base import *
from ._mesh import *
from ._path import *
from ._explicit import *

__all__ = _base.__all__ + _mesh.__all__ + _path.__all__ + _explicit.__all__



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines free functions to serialize / deserialize bands-inspect objects to HDF5.
"""

from fsc.hdf5_io import save, load, from_hdf5, from_hdf5_file, to_hdf5, to_hdf5_file

__all__ = ["save", "load", "from_hdf5", "from_hdf5_file", "to_hdf5", "to_hdf5_file"]



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines a crystal lattice class.
"""

import numpy as np
import scipy.linalg as la
from fsc.export import export

# TODO: move to a separate module  # pylint: disable=fixme,useless-suppression


@export
class Lattice:
    """
    Defines a periodic lattice.
    """

    def __init__(self, matrix):
        self.matrix = np.array(matrix)

    def __array__(self):
        return self.matrix

    @property
    def reciprocal_lattice(self):
        return type(self)(matrix=2 * np.pi * la.inv(self.matrix).T)

    def get_cartesian_coords(self, fractional_coords):
        return np.dot(fractional_coords, self.matrix)

    def get_fractional_coords(self, cartesian_coords):
        return la.solve(self.matrix.T, np.array(cartesian_coords).T).T

    def get_cartesian_distance(self, fractional_coord_1, fractional_coord_2):
        return la.norm(
            self.get_cartesian_coords(fractional_coord_1)
            - self.get_cartesian_coords(fractional_coord_2)
        )

    def get_reciprocal_cartesian_distance(  # pylint: disable=invalid-name
        self, reciprocal_fractional_coord_1, reciprocal_fractional_coord_2
    ):
        return self.reciprocal_lattice.get_cartesian_distance(
            reciprocal_fractional_coord_1, reciprocal_fractional_coord_2
        )



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines functions to plot bandstructures.
"""

from types import MappingProxyType

import decorator

from .kpoints._path import _KpointLabel


@decorator.decorator
def _plot(  # pylint: disable=missing-docstring,invalid-name,inconsistent-return-statements
    func, data, *, ax=None, **kwargs
):
    import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel

    if ax is None:
        return_fig = True
        fig, ax = plt.subplots()  # pylint: disable=invalid-name
    else:
        return_fig = False

    func(data, ax=ax, **kwargs)

    if return_fig:
        return fig


@_plot
def eigenvals(
    eigenvals,  # pylint:disable=redefined-outer-name
    *,
    ax=None,  # pylint:disable=invalid-name
    ylim=None,
    e_fermi=0.0,
    vertex_labels=True,
    energy_labels=True,
    plot_options=MappingProxyType({"color": "C0", "lw": 0.8})
):
    """
    Plot the bandstructure of a given :class:`.EigenvalsData` object.
    """
    ax.plot(eigenvals.eigenvals - e_fermi, **plot_options)
    ax.set_xlim(0, len(eigenvals.eigenvals) - 1)
    if ylim is not None:
        ax.set_ylim(ylim)
    if vertex_labels and hasattr(eigenvals.kpoints, "labels"):
        labels = _merge_labels(eigenvals.kpoints.labels)
        ax.set_xticks([l.index for l in labels])
        ax.set_xticklabels([l.label for l in labels])
        ax.xaxis.grid(True)
    else:
        ax.set_xticks([])
    if energy_labels:
        ax.yaxis.set_tick_params(direction="out", labelsize=8)
        ax.set_ylabel(r"$E$ [eV]")
    else:
        ax.set_yticks([])


def _merge_labels(labels):
    """
    Merge labels of neighbouring k-points, separating them by a ' | '.
    """
    new_labels = [labels[0]]
    for left_label, middle_label, right_label in zip(labels, labels[1:], labels[2:]):
        if middle_label.index - left_label.index == 1:
            continue
        if right_label.index - middle_label.index == 1:
            new_labels.append(
                _KpointLabel(
                    index=middle_label.index + 0.5,
                    label=middle_label.label + " | " + right_label.label,
                )
            )
        else:
            new_labels.append(middle_label)
    new_labels.append(labels[-1])
    return new_labels



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines functions to compare different bandstructures.
"""

from . import difference
from . import align



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines the data class for explicitly specified k-points.
"""

import types

import numpy as np
from fsc.export import export
from fsc.hdf5_io import subscribe_hdf5

from ._base import KpointsBase


@export
@subscribe_hdf5("bands_inspect.kpoints_explicit", extra_tags=("kpoints_explicit",))
class KpointsExplicit(KpointsBase, types.SimpleNamespace):
    """
    Defines an explicit set of k-points.

    :param kpoints: List of explicit k-points.
    :type kpoints: list
    """

    def __init__(self, kpoints):
        self.kpoints = np.array(kpoints)

    @property
    def kpoints_explicit(self):
        return self.kpoints

    def to_hdf5(self, hdf5_handle):
        hdf5_handle["kpoints"] = self.kpoints

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return cls(kpoints=hdf5_handle["kpoints"][()])



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines the bands-inspect CLI.
"""

import click
import numpy as np
import matplotlib.pyplot as plt

from . import io
from . import plot
from .compare import difference as _diff
from .compare import align as _align


@click.group()
def cli():
    pass


@cli.command()
@click.argument("eigenval_files", nargs=2, type=click.Path(exists=True, dir_okay=False))
@click.option("--energy-window", nargs=2, type=float, required=False)
def difference(eigenval_files, energy_window):
    """
    Calculate the difference between two bandstructures.
    """
    ev1, ev2 = (io.load(filename) for filename in eigenval_files)

    kwargs = {}
    if energy_window:
        kwargs["weight_eigenval"] = _diff.energy_window(*energy_window)

    click.echo(_diff.calculate(ev1, ev2, **kwargs))


@cli.command()
@click.option(
    "-i",
    "--input-files",
    nargs=2,
    type=click.Path(exists=True, dir_okay=False),
    default=["eigenvals1.hdf5", "eigenvals2.hdf5"],
)
@click.option(
    "-o",
    "--output-files",
    nargs=2,
    type=click.Path(exists=False, dir_okay=False),
    default=["eigenvals1_shifted.hdf5", "eigenvals2_shifted.hdf5"],
)
@click.option("--energy-window", nargs=2, type=float, required=False)
def align(input_files, output_files, energy_window):
    """
    Align two bandstructures.
    """
    ev1, ev2 = (io.load(filename) for filename in input_files)

    kwargs = {}
    if energy_window:
        kwargs["weight_eigenval"] = _diff.energy_window(*energy_window)
    res = _align.calculate(ev1, ev2, symmetric_eigenval_weights=False, **kwargs)
    io.save(res.eigenvals1_shifted, output_files[0])
    io.save(res.eigenvals2_shifted, output_files[1])

    click.echo(f"Shift:      {res.shift: }")
    click.echo(f"Difference: {res.difference: }")


@cli.command()
@click.option(
    "--input",
    "-i",
    type=click.Path(exists=True, dir_okay=False),
    default="eigenval.hdf5",
    help="File containing the input eigenvalues (in HDF5 format).",
)
@click.option("--output", "-o", type=click.Path(dir_okay=False))
@click.argument("slice_idx", nargs=-1, type=int)
def slice_bands(input, output, slice_idx):  # pylint: disable=redefined-builtin
    """
    Modify a bandstructure by selecting/re-arranging specific bands.
    """
    eigenvals = io.load(input)
    io.save(eigenvals.slice_bands(slice_idx), output)


@cli.command()
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, dir_okay=False),
    help="Output file for the plot.",
    default="plot.pdf",
)
@click.argument(
    "eigenvals_files", nargs=-1, type=click.Path(dir_okay=False), required=True
)
def plot_bands(eigenvals_files, output):
    """
    Plot one or more bandstructures which share the same set of k-points.
    """
    eigenvals_list = []
    for filename in eigenvals_files:
        eigenvals_list.append(io.load(filename))
    kpoints = eigenvals_list[0].kpoints.kpoints_explicit
    for eigenvals in eigenvals_list:
        if not np.allclose(kpoints, eigenvals.kpoints.kpoints_explicit):
            raise ValueError("K-points do not match!")
    _, axis = plt.subplots()
    for i, eigenvals in enumerate(eigenvals_list):
        plot.eigenvals(eigenvals, ax=axis, plot_options={"color": f"C{i}", "lw": 0.8})
    plt.savefig(output, bbox_inches="tight")



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines the data class for a k-point path.
"""

from types import MappingProxyType
from collections import namedtuple

import numpy as np
from fsc.export import export
from fsc.hdf5_io import subscribe_hdf5, to_hdf5, from_hdf5

from ..lattice import Lattice
from ..io._legacy import _hdf5_utils
from ._base import KpointsBase


@export
@subscribe_hdf5("bands_inspect.kpoints_path", extra_tags=("kpoints_path",))
class KpointsPath(KpointsBase):
    """
    Defines a k-point path.

    :param paths: List of paths. Each path is a list of special points.
    :type paths: list

    :param special_points: Mapping of special k-points. The key is the identifier of the special k-point, and the value is its position in reduced coordinates.
    :type special_points: dict

    :param kpoint_distance: Approximate distance between two neighbouring k-points along the path.
    :type kpoint_distance: float

    :param unit_cell: Unit cell of the material. The basis vectors are given as rows in a matrix.
    :type unit_cell: numpy.ndarray
    """

    def __init__(  # pylint: disable=missing-function-docstring
        self,
        *,
        paths,
        special_points=MappingProxyType({}),
        kpoint_distance=1e-3,
        unit_cell="auto",
    ):
        self._paths = [
            [_Vertex(pt, special_points) for pt in single_path] for single_path in paths
        ]
        dimensions = {pt.dimension for single_path in self._paths for pt in single_path}
        if len(dimensions) != 1:
            raise ValueError(f"Inconsistent dimensions: {dimensions}")
        dim = dimensions.pop()
        if isinstance(unit_cell, str) and unit_cell == "auto":
            uc = np.eye(dim)  # pylint: disable=invalid-name
        else:
            uc = np.array(unit_cell)  # pylint: disable=invalid-name
        if uc.shape != (dim, dim):
            raise ValueError(
                f"Inconsistent shape of the unit cell: {uc.shape}, should be {(dim, dim)}"
            )
        self._lattice = Lattice(matrix=uc)
        self._kpoint_distance = kpoint_distance
        self._evaluate_paths()

    def _evaluate_paths(self):
        """
        Evaluate all paths and calculate the corresponding explicit k-points.
        """
        self._kpoints_explicit = []
        self._labels = []
        for single_path in self._paths:
            self._evaluate_single_path(single_path)
        self._kpoints_explicit = np.array(self._kpoints_explicit)
        self._kpoints_explicit.flags.writeable = False

    def _evaluate_single_path(self, single_path):
        """
        Evaluate a single path (connected line of k-points), and calculate the explicit k-points.
        """
        num_kpoints_initial = len(self._kpoints_explicit)
        self._kpoints_explicit.append(single_path[0].frac)
        self._labels.append(
            _KpointLabel(index=num_kpoints_initial, label=single_path[0].label)
        )
        for start, end in zip(single_path, single_path[1:]):
            self._evaluate_line(start, end)

    def _evaluate_line(self, start, end):
        """
        Calculate the explicit k-points between two vertices which are part of the path.
        """
        dist = self._lattice.get_reciprocal_cartesian_distance(start.frac, end.frac)
        npoints = max(int(np.round_(dist / self._kpoint_distance)) + 1, 2)
        steps = np.linspace(0, 1, npoints)[1:]
        kpoints = [(1 - s) * start.frac + s * end.frac for s in steps]
        self._kpoints_explicit.extend(kpoints)
        self._labels.append(
            _KpointLabel(index=len(self._kpoints_explicit) - 1, label=end.label)
        )

    @property
    def kpoints_explicit(self):
        return self._kpoints_explicit

    @property
    def labels(self):
        return self._labels

    def to_hdf5(self, hdf5_handle):
        path_labels = [[pt.label for pt in single_path] for single_path in self._paths]
        to_hdf5(path_labels, hdf5_handle.create_group("path_labels"))
        special_points = {
            pt.label: pt.frac for special_path in self._paths for pt in special_path
        }
        to_hdf5(special_points, hdf5_handle.create_group("special_points"))
        hdf5_handle["kpoint_distance"] = self._kpoint_distance
        hdf5_handle["unit_cell"] = self._lattice.matrix

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        try:
            path_labels = from_hdf5(hdf5_handle["path_labels"])
            special_points = from_hdf5(hdf5_handle["special_points"])
        except ValueError:
            path_labels = _hdf5_utils.nested_list_from_hdf5(hdf5_handle["path_labels"])
            special_points = _hdf5_utils.dict_from_hdf5(hdf5_handle["special_points"])
        kpoint_distance = hdf5_handle["kpoint_distance"][()]
        unit_cell = hdf5_handle["unit_cell"][()]

        return cls(
            paths=path_labels,
            special_points=special_points,
            kpoint_distance=kpoint_distance,
            unit_cell=unit_cell,
        )


class _Vertex:
    """
    Defines a vertex in the k-point path.
    """

    def __init__(self, point, special_points):
        self.frac = np.array(special_points.get(point, point))
        self.label = str(point)

    @property
    def dimension(self):
        return self.frac.shape[0]


_KpointLabel = namedtuple("_KpointLabel", ["index", "label"])



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
This module contains functions to save and load objects, using the HDF5 format.
"""

from ._save_load import *



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines the base class for k-point data classes.
"""

import abc

from fsc.export import export
from fsc.hdf5_io import HDF5Enabled


@export
class KpointsBase(HDF5Enabled):
    """
    Base class for classes defining sets of k-points.
    """

    @property
    @abc.abstractmethod
    def kpoints_explicit(self):
        """
        Array containing all k-points explicitly.
        """

    def __iter__(self):
        return iter(self.kpoints_explicit)

    def __array__(self):
        return self.kpoints_explicit



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Helper functions for serializing objects to HDF5.
"""


def nested_list_from_hdf5(hdf5_handle):
    res = []
    for idx in sorted(hdf5_handle, key=int):
        res.append(list(hdf5_handle[idx][()]))
    return _decode_bytes_nested(res)


def _decode_bytes_nested(val):  # pylint: disable=missing-function-docstring
    if isinstance(val, bytes):
        return val.decode("utf-8")
    if isinstance(val, list):
        return [_decode_bytes_nested(v) for v in val]
    return val


def dict_from_hdf5(hdf5_handle):
    res = {}
    for key in hdf5_handle:
        res[key] = hdf5_handle[key][()]
    return res



# (c) 2017-2019, ETH Zurich, Institut fuer Theoretische Physik
# Author: Dominik Gresch <mail@greschd.ch>
"""
Defines the data container for eigenvalue data (bandstructures).
"""

import types

import numpy as np
from fsc.export import export
from fsc.hdf5_io import HDF5Enabled, subscribe_hdf5

from .kpoints import KpointsExplicit, KpointsBase
from .io import from_hdf5


@export
@subscribe_hdf5("bands_inspect.eigenvals_data", extra_tags=("eigenvals_data",))
class EigenvalsData(HDF5Enabled, types.SimpleNamespace):
    """
    Data container for the eigenvalues at a given set of k-points. The eigenvalues are automatically sorted by value.

    :param kpoints: List of k-points where the eigenvalues are given.
    :type kpoints: list

    :param eigenvals: Eigenvalues at each k-point. The outer axis corresponds to the different k-points, and the inner axis corresponds to the different eigenvalues at a given k-point.
    :type eigenvals: 2D array
    """

    def __init__(
        self, *, kpoints, eigenvals
    ):  # pylint: disable=missing-function-docstring
        if not isinstance(kpoints, KpointsBase):
            kpoints = KpointsExplicit(kpoints)

        eigenvals = np.sort(eigenvals)
        if len(kpoints.kpoints_explicit) != len(eigenvals):
            raise ValueError(
                f"Number of kpoints ({len(kpoints.kpoints_explicit)}) does not match the number of eigenvalue lists ({len(eigenvals)})"
            )
        self.kpoints = kpoints
        self.eigenvals = eigenvals

    def slice_bands(self, band_idx):
        """
        Returns a new instance which contains only the bands given in the index.

        :param band_idx: Indices for the bands in the new instance.
        :type band_idx: list
        """
        new_eigenvals = self.eigenvals.T[sorted(band_idx)].T
        return type(self)(kpoints=self.kpoints, eigenvals=new_eigenvals)

    @classmethod
    def from_eigenval_function(cls, *, kpoints, eigenval_function, listable=False):
        """
        Create an instance using a function that calculates the eigenvalues.

        :param kpoints: k-points for which the eigenvalues are to be calculated.
        :type kpoints: KpointsBase

        :param eigenval_function: Function which calculates the eigenvalues.

        :param listable: Flag showing whether the function can handle a list of k-points (``True``) or only single k-points (``False``).
        :type listable: bool
        """
        if listable:
            eigenvals = eigenval_function(kpoints.kpoints_explicit)
        else:
            eigenvals = [eigenval_function(k) for k in kpoints.kpoints_explicit]
        return cls(kpoints=kpoints, eigenvals=eigenvals)

    def to_hdf5(self, hdf5_handle):
        hdf5_handle.create_group("kpoints_obj")
        self.kpoints.to_hdf5(hdf5_handle["kpoints_obj"])
        hdf5_handle["eigenvals"] = self.eigenvals

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        kpoints = from_hdf5(hdf5_handle["kpoints_obj"])
        eigenvals = hdf5_handle["eigenvals"][()]
        return cls(kpoints=kpoints, eigenvals=eigenvals)

    def shift(self, value):
        """
        Returns an instance with eigenvalues shifted by the given value.

        :param value: The value by which the eigenvalues are shifted.
        :type value: float
        """
        new_eigenvals = self.eigenvals + value
        return type(self)(kpoints=self.kpoints, eigenvals=new_eigenvals)
